#%% ---------------------------------------------------
# Main program for paper "Community-Augmented Local-Link Intensity: a score for
# anomaly detection in graphs", by T.A. Badea, B.Dumitrescu

# Theodor-Adrian Badea 17.03.2023

# Copyright: You can use the programs at will, with no guarantee whatsoever.
# Please cite the paper if you use these programs.
#%% ---------------------------------------------------
import community
import networkx as nx
import numpy as np
import pandas as pd
import time
from scipy import stats

# check Setup section and configure as desired

def communityDetectionLouvain(G):
    G_undirected =  nx.to_undirected(G);
    partition = community.best_partition(G_undirected)
    communities = partition.values()
    communities = set(communities)
    nbCom = len(communities)
    nodeCommunityMapping = [-1 for x in range(G.number_of_nodes())]
    for node, com  in partition.items():
        nodeCommunityMapping[node] = com
    return nodeCommunityMapping

def CALLI(G):
    print("\nComputing CALLI score:")
    totalStartTime = time.time()
    print("     >> Step 1: Running Louvain community detection...")
    startTime = time.time()
    nodeCommunityMapping = communityDetectionLouvain(G)
    stopTime = time.time()
    print("     >> Step 1 time: ", stopTime - startTime)

    print("     >> Step 2: Computing predecessors & successors GAW...")
    startTime = time.time()
    N = G.number_of_nodes()
    successorsGAW = np.zeros((N, 1))
    predecessorsGAW = np.zeros((N, 1))
    for x in G:
        successorsWeights = []
        predecessorsWeights = []
        successorsX = G.successors(x)
        predecessorsX = G.predecessors(x)
        for y in successorsX:
            successorsWeights.append(G[x][y]['weight'])
        for y in predecessorsX:
            predecessorsWeights.append(G[y][x]['weight'])
        if len(successorsWeights) > 0:
            successorsGAW[x] = stats.gmean(successorsWeights)
        if len(predecessorsWeights) > 0:
            predecessorsGAW[x] = stats.gmean(predecessorsWeights)
    stopTime = time.time()
    print("     >> Step 2 time: ", stopTime - startTime)

    print("     >> Step 3: Computing intensities & augmenting to community level...")
    startTime = time.time()
    comDensities = np.zeros((N, 1))
    for x in G:
        successorsX = G.successors(x)
        predecessorsX = G.predecessors(x)
        comIntensitiesX = np.zeros((N, 1))
        for y in successorsX:
            cy = nodeCommunityMapping[y]
            xyIntensity = max(predecessorsGAW[y], successorsGAW[x])
            comIntensitiesX[cy] += xyIntensity
        for y in predecessorsX:
            cy = nodeCommunityMapping[y]
            yxIntensity = max(predecessorsGAW[x], successorsGAW[y])
            comIntensitiesX[cy] += yxIntensity
        cx = nodeCommunityMapping[x]
        comDensities[x] = np.sum(comIntensitiesX) / comIntensitiesX[cx]
    stopTime = time.time()
    print("     >> Step 3 time: ", stopTime - startTime)    

    print("     >> Step 4: Computing CALLI scores...")
    startTime = time.time()
    calliScore = np.zeros((N, 1))
    for x in G:
        successorsX = G.successors(x)
        predecessorsX = G.predecessors(x)
        for y in successorsX:
            calliScore[x] += comDensities[y]
        for y in predecessorsX:
            calliScore[x] += comDensities[y]
        calliScore[x] /= comDensities[x]
    stopTime = time.time()
    print("     >> Step 4 time: ", stopTime - startTime)   
    totalStopTime = time.time()
    print("Total time:", totalStopTime - totalStartTime)

    res = pd.DataFrame(np.arange(N), columns=["id"]) # column of node IDs
    res["CALLI"] = calliScore
    return res

#%% ---------------------------------------------------
# Setup 
syntheticGraph = True # switch False for Libra Internet Bank graph
if syntheticGraph:
    graphFileName = "synth_graph_g7"
else:
    graphFileName = "Libra_bank_3months_graph"

dfG = pd.read_csv(graphFileName + '.csv')
print(dfG)

G = nx.from_pandas_edgelist(df=dfG, source='id_source', target='id_destination',
                        edge_attr=True, create_using=nx.DiGraph)
N = G.number_of_nodes()
v = nx.get_edge_attributes(G, 'cum_amount')
nx.set_edge_attributes(G, v, 'weight')
#%% ---------------------------------------------------

#%% ---------------------------------------------------
# Compute CALLI score
fileNameCachedCALLI = graphFileName + '_calli.csv'
try:
    score = pd.read_csv(fileNameCachedCALLI) # check if score has already been computed
except FileNotFoundError:
    score = CALLI(G)
    score.to_csv(fileNameCachedCALLI, index=False)
#%% ---------------------------------------------------

nodeAlerts = np.zeros((N, 1))
if syntheticGraph == False:
    nodeReports = np.zeros((N, 1))

for index, row in dfG.iterrows():
    nodeAlerts[int(row["id_source"])] += row["nr_alerts"]
    nodeAlerts[int(row["id_destination"])] += row["nr_alerts"]
    if syntheticGraph == False:
        nodeReports[int(row["id_source"])] += row["nr_reports"]
        nodeReports[int(row["id_destination"])] += row["nr_reports"]

W = np.sum(nodeAlerts)  # number of (weighted) alerts
if syntheticGraph == False:  # same, for reports
    Wr = np.sum(nodeReports)  # number of (weighted) reports

ids = np.array(score["id"])
X = score["CALLI"].values
   
iscore = np.argsort(-X)    # sort CALLI scores decreasingly
anomalyIDs = ids[iscore]   # the IDs of reported anomalies
detectedAlerts = nodeAlerts[iscore]
if syntheticGraph == False:
    detectedReports = nodeReports[iscore]

# unweighted number of anomalies detected (number of anomalous nodes)
AN_01 = np.sum([val > 0 for val in nodeAlerts[anomalyIDs[0:int(0.001*N)]]])
AN_02 = np.sum([val > 0 for val in nodeAlerts[anomalyIDs[0:int(0.002*N)]]])
AN_05 = np.sum([val > 0 for val in nodeAlerts[anomalyIDs[0:int(0.005*N)]]])
AN_1 = np.sum([val > 0 for val in nodeAlerts[anomalyIDs[0:int(0.01*N)]]])

# cumulated weights for detected anomalous nodes
W_01 = np.sum([val for val in nodeAlerts[anomalyIDs[0:int(0.001*N)]]])
W_02 = np.sum([val for val in nodeAlerts[anomalyIDs[0:int(0.002*N)]]])
W_05 = np.sum([val for val in nodeAlerts[anomalyIDs[0:int(0.005*N)]]])
W_1 = np.sum([val for val in nodeAlerts[anomalyIDs[0:int(0.01*N)]]])

# total anomalous nodes & total anomaly weight
AN = np.sum([val > 0 for val in nodeAlerts])

print("Number of anomalous nodes for alerts (AN): ", AN)
print("Detected alerts AN 0.1% 0.2% 0.5% 1%: ", AN_01, " ", AN_02, " ", AN_05, " ", AN_1)

print("Total alerts weight: ", W)
print("Detected alerts W 0.1% 0.2% 0.5% 1%: ", W_01, " ", W_02, " ", W_05, " ", W_1)

if syntheticGraph == False:
    ANr_01 = np.sum([val > 0 for val in nodeReports[anomalyIDs[0:int(0.001*N)]]])
    ANr_02 = np.sum([val > 0 for val in nodeReports[anomalyIDs[0:int(0.002*N)]]])
    ANr_05 = np.sum([val > 0 for val in nodeReports[anomalyIDs[0:int(0.005*N)]]])
    ANr_1 = np.sum([val > 0 for val in nodeReports[anomalyIDs[0:int(0.01*N)]]])

    Wr_01 = np.sum([val for val in nodeReports[anomalyIDs[0:int(0.001*N)]]])
    Wr_02 = np.sum([val for val in nodeReports[anomalyIDs[0:int(0.002*N)]]])
    Wr_05 = np.sum([val for val in nodeReports[anomalyIDs[0:int(0.005*N)]]])
    Wr_1 = np.sum([val for val in nodeReports[anomalyIDs[0:int(0.01*N)]]])

    ANr = np.sum([val > 0 for val in nodeReports])
    
    print("Number of anomalous nodes for reports (ANr): ", AN)
    print("Detected reports ANr 0.1% 0.2% 0.5% 1%: ", ANr_01, " ", ANr_02, " ", ANr_05, " ", ANr_1)

    print("Total reports weight: ", Wr)
    print("Detected reports Wr 0.1% 0.2% 0.5% 1%: ", Wr_01, " ", Wr_02, " ", Wr_05, " ", Wr_1)

# TPR
tprAlerts = np.cumsum(np.array(detectedAlerts)) / W
if syntheticGraph == False:
    tprReports = np.cumsum(np.array(detectedReports)) / Wr

#%% ---------------------------------------------------
# Display TPR results
TPR_milestones_perc = [0.1, 0.2, 0.5, 1]  # percentage in all nodes where to compute TPR values
AUC_milestones_perc = [1, 100]            # same, for AUC computation

print("\nStatistics for ALERTS")
p_perc = np.array(TPR_milestones_perc)/100*N
p_perc = p_perc.astype('int')

print("True positives detected in first", *TPR_milestones_perc, "% anomalies")
for i in range(len(p_perc)):
    print("{:.4f}".format(tprAlerts[np.array(p_perc[i]).astype('int')]), end=' ')
print("")

a_perc = np.array(AUC_milestones_perc)/100*N
a_perc = a_perc.astype('int')
print("TPR AUC in first", *AUC_milestones_perc, "% anomalies")
for i in range(len(a_perc)):
    print("{:.4f}".format(np.average(tprAlerts[:a_perc[i]])), end=' ')
print("")

import matplotlib.pyplot as plt
plt.plot(tprAlerts)
plt.title('Alerts: True positive rate')
plt.grid()
plt.show()

if syntheticGraph == False:
    print("\nSame statistics for REPORTS")

    print("True positives detected in first", *TPR_milestones_perc, "% anomalies")
    for i in range(len(p_perc)):
        print("{:.4f}".format(tprReports[np.array(p_perc[i]).astype('int')]), end=' ')
    print("")

    print("TPR AUC in first", *AUC_milestones_perc, "% anomalies")
    for i in range(len(a_perc)):
        print("{:.4f}".format(np.average(tprReports[:a_perc[i]])), end=' ')
    print("")

    import matplotlib.pyplot as plt
    plt.plot(tprReports)
    plt.title('Reports: True positive rate')
    plt.grid()
    plt.show()